package de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl;

import de.unikoblenz.fgbks.core.dmn.domain.ids.AbstractId;
import de.unikoblenz.fgbks.core.dmn.domain.ids.DefinitionId;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDefinitions;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnElement;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputData;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.lang3.Validate;
import org.camunda.bpm.model.dmn.instance.Definitions;

/**
 * Implementation of {@link VDmnDefinitions}.
 */
public class VDmnDefinitionsImpl extends AbstractVDmnElement implements VDmnDefinitions {

  protected Definitions definitions;
  protected List<VDmnDecision> dmnDecisions;
  protected List<VDmnInputData> dmnInputData;
  protected Map<AbstractId, VDmnElement> elements;

  private VDmnDefinitionsImpl() {
    super();
    dmnDecisions = new ArrayList<>();
    dmnInputData = new ArrayList<>();
    elements = new HashMap<>();
  }

  /**
   * Get a new instance of a builder.
   *
   * @return a new builder instance
   */
  public static Builder getBuilder() {
    return new VDmnDefinitionsImpl().new Builder();
  }

  @Override
  public Definitions getDefinitions() {
    return definitions;
  }

  @Override
  public List<VDmnDecision> getVDmnDecisions() {
    return Collections.unmodifiableList(dmnDecisions);
  }

  @Override
  public VDmnElement getElementById(AbstractId id) {
    return elements.get(Validate.notNull(id));
  }

  @Override
  public List<VDmnInputData> getVDmnInputData() {
    return Collections.unmodifiableList(dmnInputData);
  }

  public class Builder extends AbstractDmnElementBuilder<VDmnDefinitionsImpl> {

    public Builder withDefinitions(Definitions definitions) {
      value.definitions = definitions;
      return this;
    }

    public Builder withDefinitionId(DefinitionId definitionId) {
      value.id = definitionId;
      return this;
    }

    public Builder addVDmnDecision(VDmnDecision dmnDecision) {
      value.dmnDecisions.add(Validate.notNull(dmnDecision));
      return this;
    }

    public Builder addVDmnDecisionFromBuilder(VDmnDecisionImpl.Builder dmnDecisionBuilder) {
      addVDmnDecision(dmnDecisionBuilder.getUnbuildValue());
      dmnDecisionBuilder.withVDmnDefinition(value);
      return this;
    }

    public Builder addVDmnInputData(VDmnInputData dmnInputData) {
      value.dmnInputData.add(Validate.notNull(dmnInputData));
      return this;
    }

    public Builder addVDmnInputDataFromBuilder(VDmnInputDataImpl.Builder dmnInputDataBuilder) {
      addVDmnInputData(dmnInputDataBuilder.getUnbuildValue());
      dmnInputDataBuilder.withVDmnDefinition(value);
      return this;
    }

    @Override
    protected void validate() {
      super.validate();
      Validate.notNull(value.id);
    }

    @Override
    public VDmnDefinitionsImpl build() {
      // save all Ids and elements in the element map
      // add the definition
      value.elements.put(getId(), this.value);
      // add the input data
      value.getVDmnInputData().forEach(inData -> value.elements.put(inData.getId(), inData));
      // add all decisions
      for (VDmnDecision decision : value.getVDmnDecisions()) {
        value.elements.put(decision.getId(), decision);
        // add the decision table
        VDmnDecisionTable table = decision.getVDmnDecisionTable();
        value.elements.put(table.getId(), table);
        // add columns
        table.getVDmnInputColumns().forEach(column -> value.elements.put(column.getId(), column));
        table
            .getVDmnInputColumns()
            .forEach(column -> value.elements.put(column.getInputExpressionId(), column));
        table.getVDmnOutputColumns().forEach(column -> value.elements.put(column.getId(), column));
        // add rules
        table.getVDmnRules().forEach(rule -> value.elements.put(rule.getId(), rule));
        // add input values/enties
        table.getVDmnInputColumns().stream()
            .flatMap(column -> column.getVDmnInputValues().stream())
            .forEach(iv -> value.elements.put(iv.getId(), iv));
        table.getVDmnOutputColumns().stream()
            .flatMap(column -> column.getVDmnOutputValues().stream())
            .forEach(iv -> value.elements.put(iv.getId(), iv));
      }
      return super.build();
    }
  }
}
